import { HttpClient } from '@angular/common/http';
import { Injectable } from '@angular/core';
import { BehaviorSubject, Observable } from 'rxjs';
import { map } from 'rxjs/operators';

import { environment } from '../../environments/environment';
import { User } from '../core/user';

const loginUrl = '/token';
const refreshUrl = '/refresh';
const JWT_USER = 'JWT_USER';

@Injectable({ providedIn: 'root' })
export class AuthService {
  private currentUserSubject: BehaviorSubject<User>;
  public currentUser: Observable<User>;

  constructor(private http: HttpClient) {
    this.checkStorage();
    this.currentUser = this.currentUserSubject.asObservable();
  }

  static parseJwt(token): User {
    const base64Url = token.split('.')[1];
    const base64 = base64Url.replace(/-/g, '+').replace(/_/g, '/');
    const jsonPayload = decodeURIComponent(
      atob(base64)
        .split('')
        .map((c) => `%${`00${c.charCodeAt(0).toString(16)}`.slice(-2)}`)
        .join(''),
    );

    const decoded = JSON.parse(jsonPayload);
    return new User({
      id: decoded.userId,
      name: decoded.sub,
      lockedOut: decoded.lockedOut,
      perms: decoded.scopes,
      access_token: token,
      exp: decoded.exp,
      ver: decoded.ver,
    });
  }

  checkStorage(): User {
    const existingToken: User = JSON.parse(localStorage.getItem(JWT_USER));
    if (existingToken === null || Date.now() > existingToken.exp * 1000) {
      localStorage.removeItem(JWT_USER);
      this.currentUserSubject = new BehaviorSubject<User>(null);
      return null;
    }
    this.currentUserSubject = new BehaviorSubject<User>(existingToken);
    return existingToken;
  }

  public get user(): User {
    let val = this.currentUserSubject.value;
    if (val == null) {
      return val;
    }
    let expired = Date.now() > val.exp * 1000;
    if (expired) {
      val = this.checkStorage();
    }
    if (val == null) {
      return null;
    }
    expired = Date.now() > val.exp * 1000;
    if (expired) {
      this.logout();
    }
    return this.currentUserSubject.value;
  }

  login(username: string, password: string, otp: string) {
    const formData: FormData = new FormData();
    formData.append('username', username);
    formData.append('password', password);
    formData.append('otp', otp);
    formData.append('grant_type', 'password');
    return this.http
      .post<any>(loginUrl, formData)
      .pipe(map((u) => u.access_token))
      .pipe(map((u) => AuthService.parseJwt(u)))
      .pipe(
        map((user) => {
          // Store user details and jwt token in local storage
          // to keep user logged in between page refreshes
          localStorage.setItem(JWT_USER, JSON.stringify(user));
          this.currentUserSubject.next(user);
          return user;
        }),
      );
  }

  needsRefreshing(): boolean {
    return Date.now() > (this.user.exp - environment.ACCESS_TOKEN_REFRESH_MINUTES * 60) * 1000;
  }

  logout() {
    // remove user from local storage to log user out
    localStorage.removeItem(JWT_USER);
    this.currentUserSubject.next(null);
  }

  refreshToken() {
    return this.http
      .post<any>(refreshUrl, {})
      .pipe(map((u) => u.access_token))
      .pipe(map((u) => AuthService.parseJwt(u)))
      .pipe(
        map((user) => {
          // Store user details and jwt token in local storage
          // to keep user logged in between page refreshes
          localStorage.setItem(JWT_USER, JSON.stringify(user));
          this.currentUserSubject.next(user);
          return user;
        }),
      );
  }
}
